@@ -45,7 +45,7 @@ module Gitlab
     # See http://stackoverflow.com/a/40884093/1856239 and
     # https://github.com/git/git/blob/3ad8b5bf26362ac67c9020bf8c30eee54a84f56d/cache.h#L1011-L1012
     EMPTY_TREE_ID = '4b825dc642cb6eb9a060e54bf8d69288fbee4904'.freeze
-    BLANK_SHA = '0' * 40
+    BLANK_SHA = ('0' * 40).freeze
     TAG_REF_PREFIX = "refs/tags/".freeze
     BRANCH_REF_PREFIX = "refs/heads/".freeze
 
@@ -100,13 +100,6 @@ module Gitlab
         Gitlab::Git::Version.git_version
       end
 
-      def check_namespace!(*objects)
-        expected_namespace = self.name + '::'
-        objects.each do |object|
-          raise ArgumentError, "expected object in #{expected_namespace}, got #{object}" unless object.class.name.start_with?(expected_namespace)
-        end
-      end
-
       def diff_line_code(file_path, new_line_position, old_line_position)
         "#{Digest::SHA1.hexdigest(file_path)}_#{old_line_position}_#{new_line_position}"
       end
