@@ -22,10 +22,6 @@ module Gitlab
 
       def initialize(user, new_repository)
         @user = user
-
-        # Refactoring aid
-        Gitlab::Git.check_namespace!(new_repository)
-
         @repository = new_repository
       end
 
@@ -69,26 +65,32 @@ module Gitlab
         end
       end
 
-      # Whenever `start_branch_name` is passed, if `branch_name` doesn't exist,
-      # it would be created from `start_branch_name`.
+      # Whenever `start_branch_name` or `start_sha` is passed, if `branch_name`
+      # doesn't exist, it will be created from the commit pointed to by
+      # `start_branch_name` or `start_sha`.
+      #
       # If `start_repository` is passed, and the branch doesn't exist,
       # it would try to find the commits from it instead of current repository.
       def with_branch(branch_name,
                       start_branch_name: nil,
+                      start_sha: nil,
                       start_repository: repository,
                       force: false,
                       &block)
-        Gitlab::Git.check_namespace!(start_repository)
         start_repository = RemoteRepository.new(start_repository) unless start_repository.is_a?(RemoteRepository)
 
         start_branch_name = nil if start_repository.empty?
 
-        raise ArgumentError, "Cannot find branch #{start_branch_name} in #{start_repository.relative_path}" if start_branch_name && !start_repository.branch_exists?(start_branch_name)
+        if start_branch_name.present? && !start_repository.branch_exists?(start_branch_name)
+          raise ArgumentError, "Cannot find branch '#{start_branch_name}'"
+        elsif start_sha.present? && !start_repository.commit_id(start_sha)
+          raise ArgumentError, "Cannot find commit '#{start_sha}'"
+        end
 
         update_branch_with_hooks(branch_name, force) do
           repository.with_repo_branch_commit(
             start_repository,
-            start_branch_name || branch_name,
+            start_sha.presence || start_branch_name.presence || branch_name,
             &block
           )
         end
@@ -107,18 +109,24 @@ module Gitlab
       # Returns the generated commit.
       #
       # ref - The target ref path we're committing to.
-      # from_branch - The branch we're taking the HEAD commit from.
-      def commit_ref(ref, from_branch:)
+      # from_ref - The ref we're taking the HEAD commit from.
+      def commit_ref(ref, source_sha, from_ref:)
         update_autocrlf_option
 
-        repository.write_ref(ref, from_branch.target)
+        target_sha = from_ref.target
+        repository.write_ref(ref, target_sha)
 
         # Make commit
         newrev = yield
 
-        raise Gitlab::Git::CommitError.new('Failed to create commit') unless newrev
+        unless newrev
+          error = "Failed to create merge commit for source_sha #{source_sha} and" \
+                  " target_sha #{target_sha} at #{ref}"
 
-        oldrev = from_branch.target
+          raise Gitlab::Git::CommitError.new(error)
+        end
+
+        oldrev = from_ref.target
 
         update_ref(ref, newrev, oldrev)
 
@@ -199,8 +207,10 @@ module Gitlab
 
         unless status.zero?
           Gitlab::GitLogger.error("'git update-ref' in #{repository.path}: #{output}")
+          ref_name = Gitlab::Git.branch_name(ref) || ref
+
           raise Gitlab::Git::CommitError.new(
-            "Could not update branch #{Gitlab::Git.branch_name(ref)}." \
+            "Could not update #{ref_name}." \
             " Please refresh and try again."
           )
         end
