@@ -26,6 +26,7 @@ module Gitlab
       GITALY_INTERNAL_URL = 'ssh://gitaly/internal.git'.freeze
       AUTOCRLF_VALUES = { 'true' => true, 'false' => false, 'input' => :input }.freeze
       RUGGED_KEY = :rugged_list
+      GIT_ALLOW_SHA_UPLOAD = 'uploadpack.allowAnySHA1InWant=true'.freeze
 
       NoRepository = Class.new(StandardError)
       InvalidRef = Class.new(StandardError)
@@ -267,19 +268,21 @@ module Gitlab
         tags.find { |tag| tag.name.b == name_b }
       end
 
-      def merge_to_ref(user, source_sha, branch, target_ref, message)
-        branch = find_branch(branch)
+      def merge_to_ref(user, source_sha, branch, target_ref, message, first_parent_ref)
+        ref = if first_parent_ref.present?
+                find_ref(first_parent_ref)
+              else
+                find_branch(branch)
+              end
 
-        raise InvalidRef unless branch
+        raise InvalidRef unless ref
 
-        OperationService.new(user, self).commit_ref(target_ref, from_branch: branch) do
-          our_commit = branch.target
+        OperationService.new(user, self).commit_ref(target_ref, source_sha, from_ref: ref) do
+          our_commit = ref.target
           their_commit = source_sha
 
           create_merge_commit(user, our_commit, their_commit, message)
         end
-      rescue Gitlab::Git::CommitError # when merge_index.conflicts?
-        nil
       rescue Rugged::ReferenceError, InvalidRef
         raise ArgumentError, 'Invalid merge source'
       end
@@ -316,8 +319,6 @@ module Gitlab
           start_repository: start_repository
         ) do |start_commit|
 
-          Gitlab::Git.check_namespace!(commit, start_repository)
-
           revert_tree_id = check_revert_content(commit, start_commit.sha)
           raise CreateTreeError unless revert_tree_id
 
@@ -453,10 +454,11 @@ module Gitlab
       # rubocop:disable Metrics/ParameterLists
       def multi_action(user, branch_name:, message:, actions:,
                        author_email: nil, author_name: nil,
-                       start_branch_name: nil, start_repository: self, force: false)
+                       start_branch_name: nil, start_sha: nil, start_repository: self, force: false)
         OperationService.new(user, self).with_branch(
           branch_name,
           start_branch_name: start_branch_name,
+          start_sha: start_sha,
           start_repository: start_repository,
           force: force
         ) do |start_commit|
@@ -486,45 +488,27 @@ module Gitlab
       end
       # rubocop:enable Metrics/ParameterLists
 
-      def with_repo_branch_commit(start_repository, start_branch_name)
-        Gitlab::Git.check_namespace!(start_repository)
+      def with_repo_branch_commit(start_repository, start_ref)
         start_repository = RemoteRepository.new(start_repository) unless start_repository.is_a?(RemoteRepository)
 
-        return yield nil if start_repository.empty?
-
-        if start_repository.same_repository?(self)
-          yield commit(start_branch_name)
-        else
-          start_commit_id = start_repository.commit_id(start_branch_name)
-
-          return yield nil unless start_commit_id
-
-          if branch_commit = commit(start_commit_id)
-            yield branch_commit
-          else
-            with_repo_tmp_commit(
-              start_repository, start_branch_name, start_commit_id
-            ) do |tmp_commit|
-              yield tmp_commit
-            end
-          end
+        if start_repository.empty?
+          return yield nil
+        elsif start_repository.same_repository?(self)
+          # Directly return the commit from this repository
+          return yield commit(start_ref)
         end
-      end
 
-      def with_repo_tmp_commit(start_repository, start_branch_name, sha)
-        source_ref = start_branch_name
+        # Find the commit from the remote repository (this triggers an RPC)
+        commit_id = start_repository.commit_id(start_ref)
+        return yield nil unless commit_id
 
-        source_ref = "#{Gitlab::Git::BRANCH_REF_PREFIX}#{source_ref}" unless Gitlab::Git.branch_ref?(source_ref)
-
-        tmp_ref = fetch_ref(
-          start_repository,
-          source_ref: source_ref,
-          target_ref: "refs/tmp/#{SecureRandom.hex}"
-        )
-
-        yield commit(sha)
-      ensure
-        delete_refs(tmp_ref) if tmp_ref
+        if existing_commit = commit(commit_id)
+          # Commit is already present (e.g. in a fork, or through a previous fetch)
+          yield existing_commit
+        else
+          fetch_sha(start_repository, commit_id)
+          yield commit(commit_id)
+        end
       end
 
       def fetch_source_branch!(source_repository, source_branch, local_ref)
@@ -548,6 +532,14 @@ module Gitlab
         end
       end
 
+      def find_ref(name)
+        rugged_ref = rugged.references[name]
+
+        return unless rugged_ref
+
+        Gitlab::Git::Ref.new(self, rugged_ref.name, rugged_ref.target, rugged_ref.target_id)
+      end
+
       # Delete the specified branch from the repository
       def delete_branch(branch_name)
         rugged.branches.delete(branch_name)
@@ -605,15 +597,17 @@ module Gitlab
         end
       end
 
-      def fetch_ref(source_repository, source_ref:, target_ref:)
-        Gitlab::Git.check_namespace!(source_repository)
+      # Fetch a commit from the given source repository
+      def fetch_sha(source_repository, sha)
         source_repository = RemoteRepository.new(source_repository) unless source_repository.is_a?(RemoteRepository)
 
-        args = %W[fetch --no-tags -f #{GITALY_INTERNAL_URL} #{source_ref}:#{target_ref}]
-        message, status = run_git(args, env: source_repository.fetch_env, include_stderr: true)
-        raise Gitlab::Git::CommandError, message if status != 0
+        env = source_repository.fetch_env(git_config_options: [GIT_ALLOW_SHA_UPLOAD])
 
-        target_ref
+        args = %W[fetch --no-tags #{GITALY_INTERNAL_URL} #{sha}]
+        message, status = run_git(args, env: env, include_stderr: true)
+        raise Gitlab::Git::CommandError, message unless status.zero?
+
+        sha
       end
 
       # Lookup for rugged object by oid or ref name
@@ -823,8 +817,6 @@ module Gitlab
           start_repository: start_repository
         ) do |start_commit|
 
-          Gitlab::Git.check_namespace!(commit, start_repository)
-
           cherry_pick_tree_id = check_cherry_pick_content(commit, start_commit.sha)
           raise CreateTreeError unless cherry_pick_tree_id
 
