@@ -227,6 +227,91 @@ describe Gitlab::Git::Repository do # rubocop:disable Metrics/BlockLength
     end
   end
 
+  describe '#with_repo_branch_commit' do
+    let(:start_repository) { Gitlab::Git::RemoteRepository.new(source_repository) }
+    let(:start_commit) { source_repository.commit }
+
+    context 'when start_repository is empty' do
+      let(:source_repository) { gitlab_git_from_gitaly(new_empty_test_repo) }
+
+      before do
+        expect(start_repository).not_to receive(:commit_id)
+        expect(repository).not_to receive(:fetch_sha)
+      end
+
+      it 'yields nil' do
+        expect do |block|
+          repository.with_repo_branch_commit(start_repository, 'master', &block)
+        end.to yield_with_args(nil)
+      end
+    end
+
+    context 'when start_repository is the same repository' do
+      let(:source_repository) { repository }
+
+      before do
+        expect(start_repository).not_to receive(:commit_id)
+        expect(repository).not_to receive(:fetch_sha)
+      end
+
+      it 'yields the commit for the SHA' do
+        expect do |block|
+          repository.with_repo_branch_commit(start_repository, start_commit.sha, &block)
+        end.to yield_with_args(start_commit)
+      end
+
+      it 'yields the commit for the branch' do
+        expect do |block|
+          repository.with_repo_branch_commit(start_repository, 'master', &block)
+        end.to yield_with_args(start_commit)
+      end
+    end
+
+    context 'when start_repository is different' do
+      let(:source_repository) { gitlab_git_from_gitaly(test_repo_read_only) }
+
+      context 'when start commit already exists' do
+        let(:start_commit) { repository.commit }
+
+        before do
+          expect(start_repository).to receive(:commit_id).and_return(start_commit.sha)
+          expect(repository).not_to receive(:fetch_sha)
+        end
+
+        it 'yields the commit for the SHA' do
+          expect do |block|
+            repository.with_repo_branch_commit(start_repository, start_commit.sha, &block)
+          end.to yield_with_args(start_commit)
+        end
+
+        it 'yields the commit for the branch' do
+          expect do |block|
+            repository.with_repo_branch_commit(start_repository, 'master', &block)
+          end.to yield_with_args(start_commit)
+        end
+      end
+
+      context 'when start commit does not exist' do
+        before do
+          expect(start_repository).to receive(:commit_id).and_return(start_commit.sha)
+          expect(repository).to receive(:fetch_sha).with(start_repository, start_commit.sha)
+        end
+
+        it 'yields the fetched commit for the SHA' do
+          expect do |block|
+            repository.with_repo_branch_commit(start_repository, start_commit.sha, &block)
+          end.to yield_with_args(nil) # since fetch_sha is mocked
+        end
+
+        it 'yields the fetched commit for the branch' do
+          expect do |block|
+            repository.with_repo_branch_commit(start_repository, 'master', &block)
+          end.to yield_with_args(nil) # since fetch_sha is mocked
+        end
+      end
+    end
+  end
+
   describe '#fetch_source_branch!' do
     let(:local_ref) { 'refs/merge-requests/1/head' }
     let(:repository) { mutable_repository }
@@ -321,6 +406,36 @@ describe Gitlab::Git::Repository do # rubocop:disable Metrics/BlockLength
     end
   end
 
+  describe '#fetch_sha' do
+    let(:source_repository) { Gitlab::Git::RemoteRepository.new(repository) }
+    let(:sha) { 'b971194ee2d047f24cb897b6fb0d7ae99c8dd0ca' }
+    let(:git_args) { %W[fetch --no-tags ssh://gitaly/internal.git #{sha}] }
+
+    before do
+      expect(source_repository).to receive(:fetch_env)
+        .with(git_config_options: ['uploadpack.allowAnySHA1InWant=true'])
+        .and_return({})
+    end
+
+    it 'fetches the commit from the source repository' do
+      expect(repository).to receive(:run_git)
+        .with(git_args, env: {}, include_stderr: true)
+        .and_return(['success', 0])
+
+      expect(repository.fetch_sha(source_repository, sha)).to eq(sha)
+    end
+
+    it 'raises an error if the commit does not exist in the source repository' do
+      expect(repository).to receive(:run_git)
+        .with(git_args, env: {}, include_stderr: true)
+        .and_return(['error', 1])
+
+      expect do
+        repository.fetch_sha(source_repository, sha)
+      end.to raise_error(Gitlab::Git::CommandError, 'error')
+    end
+  end
+
   describe '#write_config' do
     before do
       repository_rugged.config["gitlab.fullpath"] = repository_path
@@ -389,7 +504,10 @@ describe Gitlab::Git::Repository do # rubocop:disable Metrics/BlockLength
     let(:branch_head) { '6d394385cf567f80a8fd85055db1ab4c5295806f' }
     let(:source_sha) { 'cfe32cf61b73a0d5e9f13e774abde7ff789b1660' }
     let(:branch) { 'test-master' }
+    let(:first_parent_ref) { 'refs/heads/test-master' }
     let(:target_ref) { 'refs/merge-requests/999/merge' }
+    let(:arg_branch) {}
+    let(:arg_first_parent_ref) { first_parent_ref }
 
     before do
       create_branch(repository, branch, branch_head)
@@ -399,20 +517,41 @@ describe Gitlab::Git::Repository do # rubocop:disable Metrics/BlockLength
       repository.rugged.references[target_ref]
     end
 
-    it 'changes target ref to a merge between source SHA and branch' do
-      expect(fetch_target_ref).to be_nil
+    shared_examples_for 'correct behavior' do
+      it 'changes target ref to a merge between source SHA and branch' do
+        expect(fetch_target_ref).to be_nil
 
-      merge_commit_id = repository.merge_to_ref(user, source_sha, branch, target_ref, 'foo')
+        merge_commit_id = repository.merge_to_ref(user, source_sha, arg_branch, target_ref, 'foo', arg_first_parent_ref)
 
-      ref = fetch_target_ref
+        ref = fetch_target_ref
 
-      expect(ref.target.oid).to eq(merge_commit_id)
+        expect(ref.target.oid).to eq(merge_commit_id)
+      end
+
+      it 'does not change the branch HEAD' do
+        expect { repository.merge_to_ref(user, source_sha, arg_branch, target_ref, 'foo', arg_first_parent_ref) }
+          .not_to change { repository.find_ref(first_parent_ref).target }
+          .from(branch_head)
+      end
     end
 
-    it 'does not change the branch HEAD' do
-      expect { repository.merge_to_ref(user, source_sha, branch, target_ref, 'foo') }
-        .not_to change { repository.find_branch(branch).target }
-        .from(branch_head)
+    it_behaves_like 'correct behavior'
+
+    context 'when legacy branch parameter is specified and ref path is empty' do
+      it_behaves_like 'correct behavior' do
+        let(:arg_branch) { branch }
+        let(:arg_first_parent_ref) {}
+      end
+    end
+
+    context 'when conflicts detected' do
+      it 'raises Gitlab::Git::CommitError' do
+        allow(repository.rugged).to receive_message_chain(:merge_commits, :conflicts?) { true }
+
+        expect { repository.merge_to_ref(user, source_sha, arg_branch, target_ref, 'foo', arg_first_parent_ref) }
+          .to raise_error(Gitlab::Git::CommitError, "Failed to create merge commit for source_sha #{source_sha} and" \
+                                                    " target_sha #{branch_head} at #{target_ref}")
+      end
     end
   end
 
